import java.nio.MappedByteBuffer;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.DoubleBuffer;
import java.nio.IntBuffer;
import java.nio.channels.FileChannel;
import java.io.FileOutputStream;
import java.io.RandomAccessFile;

import java.io.BufferedReader;
import java.io.InputStreamReader;

import java.util.Map.Entry;
import java.awt.geom.Point2D;

//import java.util.List;
import java.util.*;
import java.util.Collection;
import org.openstreetmap.josm.data.osm.QuadBuckets;
import org.openstreetmap.josm.data.osm.Node;
import org.openstreetmap.josm.data.osm.Way;
import org.openstreetmap.josm.data.osm.OsmPrimitive;
import org.openstreetmap.josm.data.coor.QuadTiling;
import org.openstreetmap.josm.data.coor.LatLon;

class QBMain
{
    static void abort(String s)
    {
        out(s);
        Object o = null;
        o.hashCode();
    }
    static void outp(String s)
    {
        System.out.print(s);
    }
    static void out(String s)
    {
        System.out.println(s);
    }
    // periodic output
    static long last_out = -1;
    static void pout(String s)
    {
        long now = System.currentTimeMillis();
        if (now - last_out < 300)
            return;
        last_out = now;
        System.out.print(s + "\r");
    }
    static void pout(String s, int i, int total)
    {
        long now = System.currentTimeMillis();
        if ((now - last_out < 300) &&
            ((i+1) < total))
            return;
        last_out = now;
        // cast to float to keep the output size down
        System.out.print(s + " " + (float)((i+1)*100.0/total) + "% done    \r");
    }
    static boolean debug = false;
    static abstract class TimeMe {
        String desc;
        abstract int go();
    }
    static class TimeIteration extends TimeMe {
        Collection<?> c;
        TimeIteration(Collection<?> c) {
            this.c = c;
            this.desc = objname(c) + "<" + objname(peek(c)) + "> iteration";
        }
        int go() {
            int ret = -1;
            for (Object o : c ) {
                ret += o.hashCode();
            }
            return ret;
        }
    }
    static void time_iter(Collection<?> c)
    {
        time_it(new TimeIteration(c));
    }
    static Object peek(Collection<?> c)
    {
        for (Object o : c)
            return o;
        return null;
    }
    static String objname(Object o)
    {
        return o.getClass().getSimpleName();
    } 
    static double search_radius = 0.0001;
    static class TimeSearches extends TimeMe {
        QuadBuckets<?> qb;
        Collection<Node> search_for;
        TimeSearches(QuadBuckets<?> qb, Collection<Node> search_for) {
            this.qb = qb;
            this.search_for = search_for;
            this.desc = objname(qb.iterator().next()) + " search";
        }
        int go() {
            for (Node c : search_for) {
                Collection<?> tmp1 = qb.search(c, search_radius);
            }
            return -1;
        }
    }
    static void time_search(QuadBuckets<?> qb, Collection<Node> search_for)
    {
        time_it(new TimeSearches(qb, search_for));
    }
    static void time_it(TimeMe tm)
    {
        String desc = tm.desc;
        long start = System.currentTimeMillis();
        long end = 0;
        int i;
        for (i = 0; i < 100000; i++) {
            pout("doing '" + desc + "' run nr: " + i);
            tm.go();
            end = System.currentTimeMillis();
            if (end - start > 2000)
                break;
        }
        i++;
        double avg = (end-start) / (1.0*i);
        desc = QuadBuckets.MAX_OBJECTS_PER_LEVEL + ": " + desc;
        out(desc + ": " + i + " runs in " + (end-start) + " ms, " + 
            "avg time: " + (float)avg + " ms");
    }
    static Node newNode(double x, double y)
    {
        LatLon coor = new LatLon(y, x);
        return new Node(coor);
    }

    static List<Integer> line_to_ints(String line)
    {
        List<Integer> ret = new ArrayList<Integer>();
        String int_strs[] = line.split(" ");
        for (String int_str : int_strs) {
            try {
                Integer i = new Integer(int_str);
                ret.add(i);
            } catch (NumberFormatException e) {
                out("bad line->"+line+"<-");
            }
        }
        return ret;
    }
    static List<Double> line_to_doubles(String line)
    {
        List<Double> ret = new ArrayList<Double>();
        String int_strs[] = line.split(" ");
        for (String int_str : int_strs) {
            try {
                Double i = new Double(int_str);
                ret.add(i);
            } catch (NumberFormatException e) {
                out("bad line->"+line+"<-");
            }
        }
        return ret;
    }
    static HashMap<Integer, Node> id_to_node;
    static Way read_way(String line)
    {
        List<Node> nodes = new ArrayList<Node>();
        for (Integer i : line_to_ints(line)) {
            Node n = id_to_node.get(i);
            if (n == null) {
                nodes = null;
                break;
            } else {
                nodes.add(n);
            }
        }
        if(nodes != null)
            return new Way(nodes);
        return null;
    }
    static String ways_bin_fname = "or_ways.bin";
    static String nodes_bin_fname = "or_coords.bin";
    static IntBuffer intMap(String fname)
    {
        try {
        RandomAccessFile oways = new RandomAccessFile(fname, "rw");
        FileChannel int_chan = oways.getChannel();
        return int_chan.map( FileChannel.MapMode.READ_WRITE, 0, Integer.MAX_VALUE-1).asIntBuffer();
        } catch (FileNotFoundException e) {
            abort("exception: " + e);
        } catch (IOException e) {
            abort("exception: " + e);
        }
        return null;
    }
    static void write_ways_binfile(BufferedReader in)
    {
        String line;
        IntBuffer map = intMap(ways_bin_fname);
        // ways are read in one-per-line.  Each nid is
        // space-separated
        try {
            while ((line = in.readLine()) != null) {
                List<Integer> way_nids = line_to_ints(line);
                for (Integer i : way_nids) {
                    map.put(i);
                }
                map.put(0);
            }
        } catch (IOException e) {
        }
        out("done writing binfile: '" + ways_bin_fname + "'");
        System.exit(0);
    }
    static DoubleBuffer doubleMap(String fname)
    {
        try {
        RandomAccessFile onodes = new RandomAccessFile(fname, "rw");
        FileChannel double_chan = onodes.getChannel();
        return double_chan.map(FileChannel.MapMode.READ_WRITE, 0, Integer.MAX_VALUE-1).asDoubleBuffer();
        } catch (FileNotFoundException e) {
            abort("exception: " + e);
        } catch (IOException e) {
            abort("exception: " + e);
        }
        return null;
    }
    static void write_nodes_binfile(BufferedReader in)
    {
        String line;
        DoubleBuffer map = doubleMap(nodes_bin_fname);
        out("writing node binary file");
        // These are formatted like this:
        // id lat lon
        // one perl line
        try {
            while ((line = in.readLine()) != null) {
                List<Double> node = line_to_doubles(line);
                if (node.size() != 3)
                    abort("badly sized line: ->"+line+"<-");
                for (Double d : node) {
                    map.put(d);
                }
            }
        } catch (IOException e) {
        }
        out("done writing binfile: '" + nodes_bin_fname + "'");
        System.exit(0);
    }
    static void read_data()
    {
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        int nr_ways = 0;
        id_to_node = new HashMap<Integer, Node>(need_to_search+need_to_insert);
        out("reading binary coords from " + nodes_bin_fname);
        int need_total = need_to_insert + need_to_search;
        DoubleBuffer map = doubleMap(nodes_bin_fname);
        for (int j = 0; j < need_total; j++) {
            int id = (int)map.get();
            double y = map.get();
            double x = map.get();
            if (consistency_testing && x > 180 || y > 90)
                abort("id: " + id + " x: " + x + " y: " + y);
            Node c = new Node(x, y);
            if (need_to_insert > 0) {
                qb_nodes.add(c);
                if (donodes)
                    nsc.add(c);
                need_to_insert--;
                if (consistency_testing && (qb_nodes.search(c.getCoor(), 0.00000001) == null)) {
                    QuadBuckets.debug = true;
                    out("cant find node I just inserted: " + c);
                    qb_nodes.search(c, 0.00000001);
                    out(j+ " phooey" + c);
                    out(" coordinate quad: " + c.quads());
                    qb_nodes.add(c);
                    System.exit(0);
                }
                inserted_nodes.add(c);
                inserted_nodes2.add(c);
                inserted_nodes3.add(c);
            } else if (need_to_search > 0) {
                need_to_search--;
                search_coords.add(c);
            } else {
                break;
            }
            id_to_node.put(id, c);
            pout("coordinate input", j, need_total);
        }
        IntBuffer imap = intMap(ways_bin_fname);
        out("reading ways from binfile: " + ways_bin_fname);
        List<Node> nodes_for_way = new ArrayList<Node>();
        while (true) {
            Integer i = imap.get();
            if (i == null) {
                out("done reading way buffer after: " + nr_ways);
            }
            if (nr_ways > 10000)
                break;
            if (i == 0) {
                if (nodes_for_way.size() == 0)
                    continue;
                Way w = new Way(nodes_for_way);
                qb_ways.add(w);
                inserted_ways.add(w);
                //qb_nodes.add(w);
                nodes_for_way.clear();
                nr_ways++;
                pout("have read " + nr_ways + " ways");
                if (nr_ways != qb_ways.size()||
                    nr_ways != qb_ways.size_iter()) {
                    QuadBuckets.debug = true;
                    qb_ways.size_iter();
                    QuadBuckets.debug = false;
                    abort("bogus size: " + nr_ways + " vs. " + qb_ways.size() + " vs. " + qb_ways.size_iter());
                }
                continue;
            }
            Node n = id_to_node.get(i);
            if (n == null) {
                //out("unable to find node id: " + i);
                nodes_for_way.clear();
                continue;
            }
            nodes_for_way.add(id_to_node.get(i));
        }
        //int wnr = 0;
        //for (Way w : qb_ways) {
        //    out("Way["+(++wnr)+"]: " + w);
        //}
        out("\nadded "+nr_ways+" ways, size: " + qb_ways.size());
        out("sizes: " + nr_ways + " vs. " + qb_ways.size() + " vs. " + qb_ways.size_iter());
        out("done adding coordinates");
        System.err.println("done adding coordinates, doing timings...");
        //write_ways_binfile(in);
        //write_nodes_binfile(in);
    }
    static int need_to_insert = 50000;
    static int need_to_search = 350000;

    static QuadBuckets<Node> qb_nodes = new QuadBuckets<Node>();
    static QuadBuckets<Way> qb_ways = new QuadBuckets<Way>();
    static QuadBuckets qb = qb_nodes;
    static Collection<Node> search_coords = new Vector<Node>();
    static Collection<Node> inserted_nodes = new ArrayList<Node>();
    static Collection<Node> inserted_nodes2 = new LinkedList<Node>();
    static Collection<Node> inserted_nodes3 = new Storage<Node>();
    static List<Way> inserted_ways = new ArrayList<Way>(need_to_insert);
    static boolean consistency_testing = false;
    static boolean donodes = false;
    static NodeSearchCache nsc = new NodeSearchCache();

    static public void main(String[] args)
    {
        long lastprint;
        long now;
        QuadBuckets.MAX_OBJECTS_PER_LEVEL = new Integer(args[0]);
        System.err.println("set QuadBuckets.MAX_OBJECTS_PER_LEVEL: " + QuadBuckets.MAX_OBJECTS_PER_LEVEL);

        read_data();
        boolean do_comparisons = false;
        boolean do_timing = true;

        //System.exit(0);
        /*
        out("WORLD_PARTS: " + QuadBuckets.WORLD_PARTS);
        out(""+(new Node(QuadTiling.tile2LatLon((newNode(   0,  90)).quadTile()))));
        out(""+(new Node(QuadTiling.tile2LatLon((newNode(   0, -90)).quadTile()))));
        out(""+(new Node(QuadTiling.tile2LatLon((newNode(   0,   0)).quadTile()))));
        out(""+(new Node(QuadTiling.tile2LatLon((newNode(   0,   1)).quadTile()))));
        out(""+(new Node(QuadTiling.tile2LatLon((newNode(0,0)).quadTile()))));
        debug = true;
        qb.add(new Node(1,1));
        qb.add(new Node(2,2));
        if (true)
            return;
        */
        int nr_iter = 25;
        long coord_total = -1;
        time_iter(qb_nodes);
        time_iter(inserted_nodes);
        time_iter(inserted_nodes2);
        time_iter(inserted_nodes3);
        //time_search(qb_nodes, search_coords);
        //time_search(qb_ways, search_coords);
        //time_iter(qb_ways, "QuadBucket ways");
        //time_iter(inserted_ways, "ArrayList ways");
        if (do_comparisons)
        for (Node c : search_coords) {
            Collection coords1 = qb_ways.search(c, search_radius);
            Collection nodes = Collections.EMPTY_LIST;
            if (donodes)
               nodes = nsc.getNear(c.x, c.y, search_radius);
            int coords1_size = 0;
            if (coords1 != null)
                coords1_size = coords1.size();
            if (coords1_size != nodes.size()) {
                out("c.len: " + coords1_size + " n.len: " + nodes.size());
                QuadBuckets.debug = true;
                //NodeSearchCache.debug = true;
                out("BAD! searching for " + c);
                try {Thread.sleep(2000);} catch (Exception e){}
                coords1 = qb.search(c, search_radius);
                if (coords1 == null)
                    coords1 = new ArrayList<OsmPrimitive>();
                nodes = nsc.getNear(c.x, c.y, search_radius);
                out("c.len: " + coords1.size() + " n.len: " + nodes.size());
                out("coors: ");
                //for (OsmPrimitive cc : coords1)
                //    out(cc + " dist: " + cc.distance(c));
                out("nodes: ");
                //for (OsmPrimitive n : nodes)
                //    out(n + " dist: " + n.distance(c));
                for (Object n : nodes) {
                    if (coords1.contains(n))
                        continue;
                    out("research for: " + c );
                    Collection research = qb.search(c, search_radius);
                    //for (OsmPrimitive ccc : research)
                    //    out(n + " dist: " + n.distance(c));
                    out("----\nwas looking for " + c);
                    break;
                }
                break;
            }
        }
    }
}
